# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

try:
    from airflow.utils.module_loading import (
        import_string,
        iter_namespace,
        qualname,
    )

except ImportError:
    from airflow.sdk.module_loading import import_string, iter_namespace, qualname

try:
    # This function was not available in Airflow 3.0/3.1 in module_loading, but it's good to keep it in the
    # same shared module - good for reuse
    from airflow.sdk._shared.module_loading import is_valid_dotpath

except ImportError:
    # TODO: Remove it when Airflow 3.2.0 is the minimum version
    def is_valid_dotpath(path: str) -> bool:
        import re

        if not isinstance(path, str):
            return False
        pattern = r"^[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*$"
        return bool(re.match(pattern, path))


__all__ = [
    "import_string",
    "qualname",
    "iter_namespace",
]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from airflow.providers.openlineage.utils.spark import (
        inject_parent_job_information_into_spark_properties,
        inject_transport_information_into_spark_properties,
    )
    from airflow.sdk import Context
try:
    from airflow.providers.openlineage.utils.spark import (
        inject_parent_job_information_into_spark_properties,
        inject_transport_information_into_spark_properties,
    )
except ImportError:

    def inject_parent_job_information_into_spark_properties(properties: dict, context: Context) -> dict:
        log.warning(
            "Could not import `airflow.providers.openlineage.plugins.macros`."
            "Skipping the injection of OpenLineage parent job information into Spark properties."
        )
        return properties

    def inject_transport_information_into_spark_properties(properties: dict, context: Context) -> dict:
        log.warning(
            "Could not import `airflow.providers.openlineage.plugins.listener`."
            "Skipping the injection of OpenLineage transport information into Spark properties."
        )
        return properties


__all__ = [
    "inject_parent_job_information_into_spark_properties",
    "inject_transport_information_into_spark_properties",
]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Airflow compatibility imports for seamless migration from Airflow 2 to Airflow 3.

This module provides lazy imports that automatically try Airflow 3 paths first,
then fall back to Airflow 2 paths, enabling code to work across both versions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.common.compat.version_compat import AIRFLOW_V_3_0_PLUS

if TYPE_CHECKING:
    import airflow.sdk.io as io  # noqa: F401
    import airflow.sdk.timezone as timezone  # noqa: F401
    from airflow.models.xcom import XCOM_RETURN_KEY as XCOM_RETURN_KEY
    from airflow.sdk import (
        DAG as DAG,
        Asset as Asset,
        AssetAlias as AssetAlias,
        AssetAll as AssetAll,
        AssetAny as AssetAny,
        BaseHook as BaseHook,
        BaseNotifier as BaseNotifier,
        BaseOperator as BaseOperator,
        BaseOperatorLink as BaseOperatorLink,
        BaseSensorOperator as BaseSensorOperator,
        Connection as Connection,
        Context as Context,
        DagRunState as DagRunState,
        EdgeModifier as EdgeModifier,
        Label as Label,
        Metadata as Metadata,
        ObjectStoragePath as ObjectStoragePath,
        Param as Param,
        PokeReturnValue as PokeReturnValue,
        TaskGroup as TaskGroup,
        TaskInstanceState as TaskInstanceState,
        TriggerRule as TriggerRule,
        Variable as Variable,
        WeightRule as WeightRule,
        XComArg as XComArg,
        chain as chain,
        chain_linear as chain_linear,
        cross_downstream as cross_downstream,
        dag as dag,
        get_current_context as get_current_context,
        get_parsing_context as get_parsing_context,
        setup as setup,
        task as task,
        task_group as task_group,
        teardown as teardown,
    )
    from airflow.sdk._shared.listeners import hookimpl as hookimpl
    from airflow.sdk._shared.observability.metrics.stats import Stats as Stats
    from airflow.sdk.bases.branch import BaseBranchOperator as BaseBranchOperator, BranchMixIn as BranchMixIn
    from airflow.sdk.bases.decorator import (
        DecoratedMappedOperator as DecoratedMappedOperator,
        DecoratedOperator as DecoratedOperator,
        KeywordParameters as KeywordParameters,
        TaskDecorator as TaskDecorator,
        determine_kwargs as determine_kwargs,
        get_unique_task_id as get_unique_task_id,
        task_decorator_factory as task_decorator_factory,
    )
    from airflow.sdk.bases.sensor import poke_mode_only as poke_mode_only
    from airflow.sdk.bases.skipmixin import SkipMixin as SkipMixin
    from airflow.sdk.configuration import conf as conf
    from airflow.sdk.definitions.context import context_merge as context_merge
    from airflow.sdk.definitions.mappedoperator import MappedOperator as MappedOperator
    from airflow.sdk.definitions.template import literal as literal
    from airflow.sdk.exceptions import (
        AirflowConfigException as AirflowConfigException,
        AirflowException as AirflowException,
        AirflowFailException as AirflowFailException,
        AirflowNotFoundException as AirflowNotFoundException,
        AirflowOptionalProviderFeatureException as AirflowOptionalProviderFeatureException,
        AirflowSensorTimeout as AirflowSensorTimeout,
        AirflowSkipException as AirflowSkipException,
        AirflowTaskTimeout as AirflowTaskTimeout,
        ParamValidationError as ParamValidationError,
        TaskDeferred as TaskDeferred,
        XComNotFound as XComNotFound,
    )
    from airflow.sdk.lineage import (
        HookLineage as HookLineage,
        HookLineageCollector as HookLineageCollector,
        HookLineageReader as HookLineageReader,
        NoOpCollector as NoOpCollector,
        get_hook_lineage_collector as get_hook_lineage_collector,
    )
    from airflow.sdk.listener import get_listener_manager as get_listener_manager
    from airflow.sdk.log import redact as redact
    from airflow.sdk.plugins_manager import AirflowPlugin as AirflowPlugin

    # Airflow 3-only exceptions (conditionally imported)
    if AIRFLOW_V_3_0_PLUS:
        from airflow.sdk.exceptions import (
            DagRunTriggerException as DagRunTriggerException,
            DownstreamTasksSkipped as DownstreamTasksSkipped,
        )
    from airflow.sdk.execution_time.context import (
        AIRFLOW_VAR_NAME_FORMAT_MAPPING as AIRFLOW_VAR_NAME_FORMAT_MAPPING,
        context_to_airflow_vars as context_to_airflow_vars,
    )
    from airflow.sdk.execution_time.timeout import timeout as timeout
    from airflow.sdk.execution_time.xcom import XCom as XCom
    from airflow.sdk.types import TaskInstanceKey as TaskInstanceKey


from airflow.providers.common.compat._compat_utils import create_module_getattr

# Rename map for classes that changed names between Airflow 2.x and 3.x
# Format: new_name -> (new_path, old_path, old_name)
_RENAME_MAP: dict[str, tuple[str, str, str]] = {
    # Assets: Dataset -> Asset rename in Airflow 3.0
    "Asset": ("airflow.sdk", "airflow.datasets", "Dataset"),
    "AssetAlias": ("airflow.sdk", "airflow.datasets", "DatasetAlias"),
    "AssetAll": ("airflow.sdk", "airflow.datasets", "DatasetAll"),
    "AssetAny": ("airflow.sdk", "airflow.datasets", "DatasetAny"),
}

# Airflow 3-only renames (not available in Airflow 2)
_AIRFLOW_3_ONLY_RENAMES: dict[str, tuple[str, str, str]] = {}


# Import map for classes/functions/constants
# Format: class_name -> module_path(s)
# - str: single module path (no fallback)
# - tuple[str, ...]: multiple module paths (try in order, newest first)
_IMPORT_MAP: dict[str, str | tuple[str, ...]] = {
    # ============================================================================
    # Hooks
    # ============================================================================
    "BaseHook": ("airflow.sdk", "airflow.hooks.base"),
    # ============================================================================
    # Branching
    # ============================================================================
    "BaseBranchOperator": ("airflow.sdk.bases.branch", "airflow.providers.standard.operators.branch"),
    "BranchMixIn": ("airflow.sdk.bases.branch", "airflow.providers.standard.operators.branch"),
    "SkipMixin": (
        "airflow.sdk.bases.skipmixin",
        "airflow.models.skipmixin",
        "airflow.providers.standard.utils.skipmixin",
    ),
    # ============================================================================
    # Sensors
    # ============================================================================
    "BaseSensorOperator": ("airflow.sdk", "airflow.sensors.base"),
    "PokeReturnValue": ("airflow.sdk", "airflow.sensors.base"),
    "poke_mode_only": ("airflow.sdk.bases.sensor", "airflow.sensors.base"),
    # ============================================================================
    # Operators
    # ============================================================================
    "BaseOperator": ("airflow.sdk", "airflow.models.baseoperator"),
    # ============================================================================
    # Decorators
    # ============================================================================
    "task": ("airflow.sdk", "airflow.decorators"),
    "dag": ("airflow.sdk", "airflow.decorators"),
    "task_group": ("airflow.sdk", "airflow.decorators"),
    "setup": ("airflow.sdk", "airflow.decorators"),
    "teardown": ("airflow.sdk", "airflow.decorators"),
    "TaskDecorator": ("airflow.sdk.bases.decorator", "airflow.decorators"),
    "task_decorator_factory": ("airflow.sdk.bases.decorator", "airflow.decorators.base"),
    "get_unique_task_id": ("airflow.sdk.bases.decorator", "airflow.decorators.base"),
    "KeywordParameters": ("airflow.sdk.bases.decorator", "airflow.utils.operator_helpers"),
    "determine_kwargs": ("airflow.sdk.bases.decorator", "airflow.utils.operator_helpers"),
    # ============================================================================
    # Models
    # ============================================================================
    "Connection": ("airflow.sdk", "airflow.models.connection"),
    "Variable": ("airflow.sdk", "airflow.models.variable"),
    "XCom": ("airflow.sdk.execution_time.xcom", "airflow.models.xcom"),
    "DAG": ("airflow.sdk", "airflow.models.dag"),
    "Param": ("airflow.sdk", "airflow.models.param"),
    "XComArg": ("airflow.sdk", "airflow.models.xcom_arg"),
    "DecoratedOperator": ("airflow.sdk.bases.decorator", "airflow.decorators.base"),
    "DecoratedMappedOperator": ("airflow.sdk.bases.decorator", "airflow.decorators.base"),
    "MappedOperator": ("airflow.sdk.definitions.mappedoperator", "airflow.models.mappedoperator"),
    # ============================================================================
    # Assets (Dataset → Asset rename in Airflow 3.0)
    # ============================================================================
    # Note: Asset, AssetAlias, AssetAll, AssetAny are handled by _RENAME_MAP
    # Metadata moved from airflow.datasets.metadata (2.x) to airflow.sdk (3.x)
    "Metadata": ("airflow.sdk", "airflow.datasets.metadata"),
    # ============================================================================
    # Notifiers
    # ============================================================================
    "BaseNotifier": ("airflow.sdk", "airflow.notifications.basenotifier"),
    # ============================================================================
    # Plugins
    # ============================================================================
    "AirflowPlugin": ("airflow.sdk.plugins_manager", "airflow.plugins_manager"),
    # ============================================================================
    # Operator Links & Task Groups
    # ============================================================================
    "BaseOperatorLink": ("airflow.sdk", "airflow.models.baseoperatorlink"),
    "TaskInstanceKey": ("airflow.sdk.types", "airflow.models.taskinstancekey"),
    "TaskGroup": ("airflow.sdk", "airflow.utils.task_group"),
    # ============================================================================
    # Operator Utilities (chain, cross_downstream, etc.)
    # ============================================================================
    "chain": ("airflow.sdk", "airflow.models.baseoperator"),
    "chain_linear": ("airflow.sdk", "airflow.models.baseoperator"),
    "cross_downstream": ("airflow.sdk", "airflow.models.baseoperator"),
    # ============================================================================
    # Edge Modifiers & Labels
    # ============================================================================
    "EdgeModifier": ("airflow.sdk", "airflow.utils.edgemodifier"),
    "Label": ("airflow.sdk", "airflow.utils.edgemodifier"),
    # ============================================================================
    # State Enums
    # ============================================================================
    "DagRunState": ("airflow.sdk", "airflow.utils.state"),
    "TaskInstanceState": ("airflow.sdk", "airflow.utils.state"),
    "TriggerRule": ("airflow.sdk", "airflow.utils.trigger_rule"),
    "WeightRule": ("airflow.sdk", "airflow.utils.weight_rule"),
    # ============================================================================
    # IO & Storage
    # ============================================================================
    "ObjectStoragePath": ("airflow.sdk", "airflow.io.path"),
    # ============================================================================
    # Template Utilities
    # ============================================================================
    "literal": ("airflow.sdk.definitions.template", "airflow.utils.template"),
    # ============================================================================
    # Context & Utilities
    # ============================================================================
    "Context": ("airflow.sdk", "airflow.utils.context"),
    "context_merge": ("airflow.sdk.definitions.context", "airflow.utils.context"),
    "context_to_airflow_vars": ("airflow.sdk.execution_time.context", "airflow.utils.operator_helpers"),
    "AIRFLOW_VAR_NAME_FORMAT_MAPPING": (
        "airflow.sdk.execution_time.context",
        "airflow.utils.operator_helpers",
    ),
    "get_current_context": ("airflow.sdk", "airflow.operators.python"),
    "get_parsing_context": ("airflow.sdk", "airflow.utils.dag_parsing_context"),
    # ============================================================================
    # Timeout Utilities
    # ============================================================================
    "timeout": ("airflow.sdk.execution_time.timeout", "airflow.utils.timeout"),
    # ============================================================================
    # XCom & Task Communication
    # ============================================================================
    "XCOM_RETURN_KEY": "airflow.models.xcom",
    # ============================================================================
    # Lineage
    # ============================================================================
    "HookLineageCollector": ("airflow.sdk.lineage", "airflow.lineage.hook"),
    "HookLineageReader": ("airflow.sdk.lineage", "airflow.lineage.hook"),
    "get_hook_lineage_collector": ("airflow.sdk.lineage", "airflow.lineage.hook"),
    "HookLineage": ("airflow.sdk.lineage", "airflow.lineage.hook"),
    # Note: AssetLineageInfo is handled by _RENAME_MAP (DatasetLineageInfo -> AssetLineageInfo)
    "NoOpCollector": ("airflow.sdk.lineage", "airflow.lineage.hook"),
    # ============================================================================
    # Exceptions (deprecated in airflow.exceptions, prefer SDK)
    # ============================================================================
    # Note: AirflowException and AirflowNotFoundException are not deprecated, but exposing them
    # here keeps provider imports consistent across Airflow 2 and 3.
    "AirflowException": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "AirflowFailException": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "AirflowNotFoundException": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "AirflowOptionalProviderFeatureException": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "AirflowSkipException": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "AirflowTaskTimeout": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "AirflowSensorTimeout": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "ParamValidationError": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "TaskDeferred": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "XComNotFound": ("airflow.sdk.exceptions", "airflow.exceptions"),
    # ============================================================================
    # Observability
    # ============================================================================
    "Stats": ("airflow.sdk.observability.stats", "airflow.observability.stats", "airflow.stats"),
    # ============================================================================
    # Secrets Masking
    # ============================================================================
    "redact": (
        "airflow.sdk.log",
        "airflow.sdk._shared.secrets_masker",
        "airflow.sdk.execution_time.secrets_masker",
        "airflow.utils.log.secrets_masker",
    ),
    # ============================================================================
    # Listeners
    # ============================================================================
    "hookimpl": ("airflow.sdk._shared.listeners", "airflow.listeners"),
    "get_listener_manager": ("airflow.sdk.listener", "airflow.listeners.listener"),
    # Configuration
    # ============================================================================
    "conf": ("airflow.sdk.configuration", "airflow.configuration"),
    "AirflowConfigException": ("airflow.sdk.exceptions", "airflow.exceptions"),
}

# Airflow 3-only exceptions (not available in Airflow 2)
_AIRFLOW_3_ONLY_EXCEPTIONS: dict[str, tuple[str, ...]] = {
    "DownstreamTasksSkipped": ("airflow.sdk.exceptions", "airflow.exceptions"),
    "DagRunTriggerException": ("airflow.sdk.exceptions", "airflow.exceptions"),
}

# Add Airflow 3-only exceptions and renames to _IMPORT_MAP if running Airflow 3+
if AIRFLOW_V_3_0_PLUS:
    _IMPORT_MAP.update(_AIRFLOW_3_ONLY_EXCEPTIONS)
    _RENAME_MAP.update(_AIRFLOW_3_ONLY_RENAMES)
    # AssetLineageInfo exists in 3.0+ but location changed in 3.2
    # 3.0-3.1: airflow.lineage.hook.AssetLineageInfo
    # 3.2+: airflow.sdk.lineage.AssetLineageInfo
    _IMPORT_MAP["AssetLineageInfo"] = ("airflow.sdk.lineage", "airflow.lineage.hook")

# Module map: module_name -> module_path(s)
# For entire modules that have been moved (e.g., timezone)
# Usage: from airflow.providers.common.compat.sdk import timezone
_MODULE_MAP: dict[str, str | tuple[str, ...]] = {
    "timezone": ("airflow.sdk.timezone", "airflow.utils.timezone"),
    "io": ("airflow.sdk.io", "airflow.io"),
}

# Use the shared utility to create __getattr__
__getattr__ = create_module_getattr(
    import_map=_IMPORT_MAP,
    module_map=_MODULE_MAP,
    rename_map=_RENAME_MAP,
)

__all__ = list(_RENAME_MAP.keys()) + list(_IMPORT_MAP.keys()) + list(_MODULE_MAP.keys())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.common.compat._compat_utils import create_module_getattr
from airflow.providers.common.compat.version_compat import (
    AIRFLOW_V_3_1_PLUS,
    AIRFLOW_V_3_2_PLUS,
)

_IMPORT_MAP: dict[str, str | tuple[str, ...]] = {
    # Re-export from sdk (which handles Airflow 2.x/3.x fallbacks)
    "BaseBranchOperator": "airflow.providers.common.compat.sdk",
    "BaseOperator": "airflow.providers.common.compat.sdk",
    "BaseAsyncOperator": "airflow.providers.common.compat.sdk",
    "BranchMixIn": "airflow.providers.common.compat.sdk",
    "get_current_context": "airflow.providers.common.compat.sdk",
    "is_async_callable": "airflow.providers.common.compat.sdk",
    # Standard provider items with direct fallbacks
    "PythonOperator": ("airflow.providers.standard.operators.python", "airflow.operators.python"),
    "ShortCircuitOperator": ("airflow.providers.standard.operators.python", "airflow.operators.python"),
    "_SERIALIZERS": ("airflow.providers.standard.operators.python", "airflow.operators.python"),
}

if TYPE_CHECKING:
    from airflow.sdk.bases.decorator import is_async_callable
    from airflow.sdk.bases.operator import BaseAsyncOperator
elif AIRFLOW_V_3_2_PLUS:
    from airflow.sdk.bases.decorator import is_async_callable
    from airflow.sdk.bases.operator import BaseAsyncOperator
else:
    if AIRFLOW_V_3_1_PLUS:
        from airflow.sdk import BaseOperator
    else:
        from airflow.models import BaseOperator

    def is_async_callable(func) -> bool:
        """Detect if a callable is an async function."""
        import inspect
        from functools import partial

        while isinstance(func, partial):
            func = func.func
        return inspect.iscoroutinefunction(func)

    class BaseAsyncOperator(BaseOperator):
        """Stub for Airflow < 3.2 that raises a clear error."""

        @property
        def is_async(self) -> bool:
            return True

        async def aexecute(self, context):
            raise NotImplementedError()

        def execute(self, context):
            raise RuntimeError(
                "Async operators require Airflow 3.2+. Upgrade Airflow or use a synchronous callable."
            )


__getattr__ = create_module_getattr(import_map=_IMPORT_MAP)

__all__ = sorted(_IMPORT_MAP.keys())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from airflow.providers.common.compat._compat_utils import create_module_getattr

_IMPORT_MAP: dict[str, str | tuple[str, ...]] = {
    "SkipMixin": (
        "airflow.sdk.bases.skipmixin",
        "airflow.models.skipmixin",
        "airflow.providers.standard.utils.skipmixin",
    ),
    "XCOM_SKIPMIXIN_KEY": (
        "airflow.sdk.bases.skipmixin",
        "airflow.providers.standard.utils.skipmixin",
    ),
    "XCOM_SKIPMIXIN_SKIPPED": (
        "airflow.sdk.bases.skipmixin",
        "airflow.providers.standard.utils.skipmixin",
    ),
    "XCOM_SKIPMIXIN_FOLLOWED": (
        "airflow.sdk.bases.skipmixin",
        "airflow.providers.standard.utils.skipmixin",
    ),
    "write_python_script": (
        "airflow.providers.standard.utils.python_virtualenv",
        "airflow.utils.python_virtualenv",
    ),
    "prepare_virtualenv": (
        "airflow.providers.standard.utils.python_virtualenv",
        "airflow.utils.python_virtualenv",
    ),
}

__getattr__ = create_module_getattr(import_map=_IMPORT_MAP)

__all__ = sorted(_IMPORT_MAP.keys())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import functools
import logging
from importlib import metadata
from typing import Any

from packaging.version import Version

from airflow.providers.common.compat.sdk import AirflowOptionalProviderFeatureException

log = logging.getLogger(__name__)


def require_openlineage_version(
    provider_min_version: str | None | Any = None, client_min_version: str | None = None
):
    """
    Enforce minimum version requirements for OpenLineage provider or client.

    Some providers, such as Snowflake and DBT Cloud, do not require an OpenLineage provider but may
    offer optional features that depend on it. These features are generally available starting
    from a specific version of the OpenLineage provider or client. This decorator helps ensure compatibility,
    preventing import errors and providing clear logs about version requirements.

    Args:
        provider_min_version: Optional minimum version requirement for apache-airflow-providers-openlineage
        client_min_version: Optional minimum version requirement for openlineage-python

    Raises:
        ValueError: If neither `provider_min_version` nor `client_min_version` is provided.
        TypeError: If the decorator is used without parentheses (e.g., `@require_openlineage_version`).
    """
    err_msg = (
        "`require_openlineage_version` decorator must be used with at least one argument: "
        "'provider_min_version' or 'client_min_version', "
        'e.g., @require_openlineage_version(provider_min_version="1.0.0")'
    )
    # Detect if decorator is mistakenly used without arguments
    if callable(provider_min_version) and client_min_version is None:
        raise TypeError(err_msg)

    # Ensure at least one argument is provided
    if provider_min_version is None and client_min_version is None:
        raise ValueError(err_msg)

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if provider_min_version:
                try:
                    provider_version: str = metadata.version("apache-airflow-providers-openlineage")
                except metadata.PackageNotFoundError:
                    try:
                        from airflow.providers.openlineage import __version__ as provider_version
                    except (ImportError, AttributeError, ModuleNotFoundError):
                        raise AirflowOptionalProviderFeatureException(
                            "OpenLineage provider not found or has no version, "
                            f"skipping function `{func.__name__}` execution"
                        )

                if provider_version and Version(provider_version) < Version(provider_min_version):
                    raise AirflowOptionalProviderFeatureException(
                        f"OpenLineage provider version `{provider_version}` "
                        f"is lower than required `{provider_min_version}`, "
                        f"skipping function `{func.__name__}` execution"
                    )

            if client_min_version:
                try:
                    client_version: str = metadata.version("openlineage-python")
                except metadata.PackageNotFoundError:
                    raise AirflowOptionalProviderFeatureException(
                        f"OpenLineage client not found, skipping function `{func.__name__}` execution"
                    )

                if client_version and Version(client_version) < Version(client_min_version):
                    raise AirflowOptionalProviderFeatureException(
                        f"OpenLineage client version `{client_version}` "
                        f"is lower than required `{client_min_version}`, "
                        f"skipping function `{func.__name__}` execution"
                    )

            return func(*args, **kwargs)

        return wrapper

    return decorator



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from airflow.providers.openlineage.utils.utils import translate_airflow_asset
else:
    try:
        from airflow.providers.openlineage.utils.utils import translate_airflow_asset
    except ImportError:
        from airflow.providers.openlineage.utils.utils import translate_airflow_dataset

        def rename_asset_as_dataset(function):
            @wraps(function)
            def wrapper(*args, **kwargs):
                if "asset" in kwargs:
                    kwargs["dataset"] = kwargs.pop("asset")
                return function(*args, **kwargs)

            return wrapper

        translate_airflow_asset = rename_asset_as_dataset(translate_airflow_dataset)


__all__ = ["translate_airflow_asset"]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.common.compat.lineage.hook import _lacks_add_extra_method, _lacks_asset_methods

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS


@pytest.fixture
def collector():
    from airflow.providers.common.compat.sdk import HookLineageCollector

    from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_2_PLUS

    hlc = HookLineageCollector()

    if AIRFLOW_V_3_0_PLUS:
        from unittest import mock

        patch_target = "airflow.lineage.hook.get_hook_lineage_collector"
        if AIRFLOW_V_3_2_PLUS:
            patch_target = "airflow.sdk.lineage.get_hook_lineage_collector"

        with mock.patch(patch_target, return_value=hlc):
            from airflow.providers.common.compat.lineage.hook import get_hook_lineage_collector

            yield get_hook_lineage_collector()
    else:
        from airflow.lineage import hook
        from airflow.providers.common.compat.lineage.hook import get_hook_lineage_collector

        hook._hook_lineage_collector = hlc
        yield get_hook_lineage_collector()
        hook._hook_lineage_collector = None


@pytest.fixture
def noop_collector():
    from airflow.providers.common.compat.sdk import NoOpCollector

    from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_2_PLUS

    noop = NoOpCollector()

    if AIRFLOW_V_3_0_PLUS:
        from unittest import mock

        patch_target = "airflow.lineage.hook.get_hook_lineage_collector"
        if AIRFLOW_V_3_2_PLUS:
            patch_target = "airflow.sdk.lineage.get_hook_lineage_collector"

        with mock.patch(patch_target, return_value=noop):
            from airflow.providers.common.compat.lineage.hook import get_hook_lineage_collector

            yield get_hook_lineage_collector()
    else:
        from airflow.lineage import hook
        from airflow.providers.common.compat.lineage.hook import get_hook_lineage_collector

        hook._hook_lineage_collector = noop
        yield get_hook_lineage_collector()
        hook._hook_lineage_collector = None


@pytest.fixture(params=["collector", "noop_collector"])
def any_collector(request):
    return request.getfixturevalue(request.param)


def test_lacks_asset_methods_all_present():
    class Collector:
        def add_input_asset(self):
            pass

        def add_output_asset(self):
            pass

        @property
        def collected_assets(self):
            return "<HookLineage object usually>"

        def create_asset(self):
            pass

    assert _lacks_asset_methods(Collector()) is False


def test_lacks_asset_methods_missing_few():
    class Collector:
        def add_input_asset(self):
            pass

        @property
        def collected_assets(self):
            return "<HookLineage object usually>"

    assert _lacks_asset_methods(Collector()) is True


def test_lacks_asset_methods_none_present():
    class Collector:
        def add_input_dataset(self):
            pass

        def add_output_dataset(self):
            pass

    assert _lacks_asset_methods(Collector()) is True


def test_lacks_add_extra_method_present():
    class Collector:
        def __init__(self):
            self._extra = {}

        def add_extra(self):
            pass

    assert _lacks_add_extra_method(Collector()) is False


def test_lacks_add_extra_method_missing():
    class Collector:
        pass

    assert _lacks_add_extra_method(Collector()) is True


def test_retrieval_does_not_raise():  # do not use fixture here
    from airflow.providers.common.compat.lineage.hook import get_hook_lineage_collector

    # On compat tests this goes into ImportError code path
    assert get_hook_lineage_collector() is not None
    assert get_hook_lineage_collector() is not None


def test_global_collector_is_reused():  # do not use fixture here
    from airflow.providers.common.compat.lineage.hook import get_hook_lineage_collector

    c1 = get_hook_lineage_collector()
    c2 = get_hook_lineage_collector()

    assert c1 is c2


def test_all_required_methods_exist(any_collector):
    """Test that all required methods exist regardless of version."""

    # Core methods that should always exist
    assert hasattr(any_collector, "add_input_asset")
    assert hasattr(any_collector, "add_output_asset")
    assert hasattr(any_collector, "add_extra")
    assert hasattr(any_collector, "collected_assets")
    assert hasattr(any_collector, "create_asset")

    # Verify they're callable
    assert callable(any_collector.add_input_asset)
    assert callable(any_collector.add_output_asset)
    assert callable(any_collector.add_extra)
    assert callable(any_collector.create_asset)


def test_empty_collector(any_collector):
    """Test that empty collector returns empty lineage."""
    lineage = any_collector.collected_assets

    assert lineage.inputs == []
    assert lineage.outputs == []
    assert lineage.extra == []


@pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow < 3.0")
def test_af2_collector_has_dataset_methods(any_collector):
    """Test that AF 2.x also has dataset methods."""

    assert hasattr(any_collector, "add_input_dataset")
    assert hasattr(any_collector, "add_output_dataset")
    assert hasattr(any_collector, "collected_datasets")
    assert hasattr(any_collector, "create_dataset")

    assert callable(any_collector.add_input_dataset)
    assert callable(any_collector.add_output_dataset)
    assert callable(any_collector.create_dataset)


@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3.0+")
def test_af3_collector_do_not_have_dataset_methods(any_collector):
    with pytest.raises(AttributeError):
        any_collector.add_input_dataset
    with pytest.raises(AttributeError):
        any_collector.add_output_dataset
    with pytest.raises(AttributeError):
        any_collector.collected_datasets
    with pytest.raises(AttributeError):
        any_collector.create_dataset


class TestCollectorAddExtra:
    def test_add_extra_basic_functionality(self, collector):
        """Test basic add_extra functionality."""
        mock_context = mock.MagicMock()
        collector.add_extra(mock_context, "test_key", {"data": "value"})

        assert collector.has_collected

        lineage = collector.collected_assets
        assert hasattr(lineage, "extra")
        assert len(lineage.extra) == 1
        assert lineage.extra[0].key == "test_key"
        assert lineage.extra[0].value == {"data": "value"}
        assert lineage.extra[0].count == 1
        assert lineage.extra[0].context == mock_context

    def test_add_extra_count_tracking(self, collector):
        """Test that duplicate extra entries are counted correctly."""
        mock_context = mock.MagicMock()

        # Add same extra multiple times
        collector.add_extra(mock_context, "test_key", {"data": "value"})
        collector.add_extra(mock_context, "test_key", {"data": "value"})
        collector.add_extra(mock_context, "test_key", {"data": "value"})

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 1
        assert lineage.extra[0].count == 3

    def test_add_extra_different_values(self, collector):
        """Test that different values are tracked separately."""
        mock_context = mock.MagicMock()

        # Add different values
        collector.add_extra(mock_context, "key1", {"data": "value1"})
        collector.add_extra(mock_context, "key2", {"data": "value2"})
        collector.add_extra(mock_context, "key1", {"data": "value3"})

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 3
        assert lineage.extra[0].key == "key1"
        assert lineage.extra[0].value == {"data": "value1"}
        assert lineage.extra[0].count == 1
        assert lineage.extra[0].context == mock_context
        assert lineage.extra[1].key == "key2"
        assert lineage.extra[1].value == {"data": "value2"}
        assert lineage.extra[1].count == 1
        assert lineage.extra[1].context == mock_context
        assert lineage.extra[2].key == "key1"
        assert lineage.extra[2].value == {"data": "value3"}
        assert lineage.extra[2].count == 1
        assert lineage.extra[2].context == mock_context

    def test_add_extra_different_contexts(self, collector):
        """Test that different contexts are tracked separately."""
        mock_context1 = mock.MagicMock()
        mock_context2 = mock.MagicMock()

        # Add same key/value with different contexts
        collector.add_extra(mock_context1, "test_key", {"data": "value"})
        collector.add_extra(mock_context2, "test_key", {"data": "value"})

        assert collector.has_collected

        lineage = collector.collected_assets
        assert lineage.extra[0].key == "test_key"
        assert lineage.extra[0].value == {"data": "value"}
        assert lineage.extra[0].count == 1
        assert lineage.extra[0].context == mock_context1
        assert lineage.extra[1].key == "test_key"
        assert lineage.extra[1].value == {"data": "value"}
        assert lineage.extra[1].count == 1
        assert lineage.extra[1].context == mock_context2

    def test_add_extra_missing_key(self, collector):
        """Test that add_extra handles missing key gracefully."""
        mock_context = mock.MagicMock()

        # Try to add with empty key
        collector.add_extra(mock_context, "", {"data": "value"})
        collector.add_extra(mock_context, None, {"data": "value"})

        assert not collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 0

    def test_add_extra_missing_value(self, collector):
        """Test that add_extra handles missing value gracefully."""
        mock_context = mock.MagicMock()

        # Try to add with empty/None value
        collector.add_extra(mock_context, "key", "")
        collector.add_extra(mock_context, "key", None)

        assert not collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 0

    def test_add_extra_max_limit(self, collector):
        """Test that add_extra respects maximum limit."""
        mock_context = mock.MagicMock()
        max_limit = 200

        # Add more than max allowed
        for i in range(max_limit + 10):
            collector.add_extra(mock_context, f"key_{i}", {"data": f"value_{i}"})

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == max_limit

    def test_add_extra_complex_values(self, collector):
        """Test that add_extra handles complex JSON-serializable values."""
        mock_context = mock.MagicMock()

        # Add various complex types
        collector.add_extra(mock_context, "dict", {"nested": {"data": "value"}})
        collector.add_extra(mock_context, "list", [1, 2, 3, "test"])
        collector.add_extra(mock_context, "number", 42)
        collector.add_extra(mock_context, "string", "simple string")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 4
        assert lineage.extra[0].key == "dict"
        assert lineage.extra[0].value == {"nested": {"data": "value"}}
        assert lineage.extra[0].count == 1
        assert lineage.extra[0].context == mock_context
        assert lineage.extra[1].key == "list"
        assert lineage.extra[1].value == [1, 2, 3, "test"]
        assert lineage.extra[1].count == 1
        assert lineage.extra[1].context == mock_context
        assert lineage.extra[2].key == "number"
        assert lineage.extra[2].value == 42
        assert lineage.extra[2].count == 1
        assert lineage.extra[2].context == mock_context
        assert lineage.extra[3].key == "string"
        assert lineage.extra[3].value == "simple string"
        assert lineage.extra[3].count == 1
        assert lineage.extra[3].context == mock_context


class TestCollectorAddAssets:
    def test_add_asset_basic_functionality(self, collector):
        """Test basic add_input_asset and add_output_asset functionality."""
        mock_context = mock.MagicMock()

        collector.add_input_asset(mock_context, uri="s3://bucket/input-file")
        collector.add_output_asset(mock_context, uri="s3://bucket/output-file")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].asset.uri == "s3://bucket/input-file"
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[0].context == mock_context

        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].asset.uri == "s3://bucket/output-file"
        assert lineage.outputs[0].count == 1
        assert lineage.outputs[0].context == mock_context

    def test_add_asset_count_tracking(self, collector):
        """Test that duplicate assets are counted correctly."""
        mock_context = mock.MagicMock()

        # Add same input multiple times
        collector.add_input_asset(mock_context, uri="s3://bucket/input")
        collector.add_input_asset(mock_context, uri="s3://bucket/input")
        collector.add_input_asset(mock_context, uri="s3://bucket/input")

        # Add same output multiple times
        collector.add_output_asset(mock_context, uri="s3://bucket/output")
        collector.add_output_asset(mock_context, uri="s3://bucket/output")

        assert collector.has_collected

        lineage = collector.collected_assets

        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].asset.uri == "s3://bucket/input"
        assert lineage.inputs[0].count == 3
        assert lineage.inputs[0].context == mock_context

        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].asset.uri == "s3://bucket/output"
        assert lineage.outputs[0].count == 2
        assert lineage.outputs[0].context == mock_context

    def test_add_asset_different_uris(self, collector):
        """Test that different URIs are tracked separately."""
        mock_context = mock.MagicMock()

        # Add different input URIs
        collector.add_input_asset(mock_context, uri="s3://bucket/file1")
        collector.add_input_asset(mock_context, uri="s3://bucket/file2")
        collector.add_input_asset(mock_context, uri="postgres://example.com:5432/database/default/table")

        # Add different output URIs
        collector.add_output_asset(mock_context, uri="s3://output/file1")
        collector.add_output_asset(mock_context, uri="s3://output/file2")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 3
        assert lineage.inputs[0].asset.uri == "s3://bucket/file1"
        assert lineage.inputs[1].asset.uri == "s3://bucket/file2"
        assert lineage.inputs[2].asset.uri == "postgres://example.com:5432/database/default/table"

        assert len(lineage.outputs) == 2
        assert lineage.outputs[0].asset.uri == "s3://output/file1"
        assert lineage.outputs[1].asset.uri == "s3://output/file2"

    def test_add_asset_different_contexts(self, collector):
        """Test that different contexts are tracked separately."""
        mock_context1 = mock.MagicMock()
        mock_context2 = mock.MagicMock()

        # Add same URI with different contexts
        collector.add_input_asset(mock_context1, uri="s3://bucket/file")
        collector.add_input_asset(mock_context2, uri="s3://bucket/file")

        collector.add_output_asset(mock_context1, uri="s3://output/file")
        collector.add_output_asset(mock_context2, uri="s3://output/file")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 2
        assert lineage.inputs[0].asset.uri == "s3://bucket/file"
        assert lineage.inputs[0].context == mock_context1
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[1].asset.uri == "s3://bucket/file"
        assert lineage.inputs[1].context == mock_context2
        assert lineage.inputs[1].count == 1

        assert len(lineage.outputs) == 2
        assert lineage.outputs[0].asset.uri == "s3://output/file"
        assert lineage.outputs[0].context == mock_context1
        assert lineage.outputs[0].count == 1
        assert lineage.outputs[1].asset.uri == "s3://output/file"
        assert lineage.outputs[1].context == mock_context2
        assert lineage.outputs[1].count == 1

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3.0+")
    def test_add_asset_with_name_and_group(self, collector):
        """Test adding assets with name and group parameters."""
        mock_context = mock.MagicMock()

        collector.add_input_asset(mock_context, uri="s3://bucket/file", name="my-input", group="input-group")
        collector.add_output_asset(
            mock_context, uri="s3://output/file", name="my-output", group="output-group"
        )

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].asset.uri == "s3://bucket/file"
        assert lineage.inputs[0].asset.name == "my-input"
        assert lineage.inputs[0].asset.group == "input-group"
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[0].context == mock_context

        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].asset.uri == "s3://output/file"
        assert lineage.outputs[0].asset.name == "my-output"
        assert lineage.outputs[0].asset.group == "output-group"
        assert lineage.outputs[0].count == 1
        assert lineage.outputs[0].context == mock_context

    def test_add_asset_with_extra_metadata(self, collector):
        """Test adding assets with extra metadata."""
        mock_context = mock.MagicMock()

        collector.add_input_asset(
            mock_context,
            uri="postgres://example.com:5432/database/default/table",
            asset_extra={"schema": "public", "table": "users"},
        )
        collector.add_output_asset(
            mock_context,
            uri="postgres://example.com:5432/database/default/table",
            asset_extra={"schema": "public", "table": "results"},
        )

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].asset.uri == "postgres://example.com:5432/database/default/table"
        assert lineage.inputs[0].asset.extra == {"schema": "public", "table": "users"}
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[0].context == mock_context

        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].asset.uri == "postgres://example.com:5432/database/default/table"
        assert lineage.outputs[0].asset.extra == {"schema": "public", "table": "results"}
        assert lineage.outputs[0].count == 1
        assert lineage.outputs[0].context == mock_context

    def test_add_asset_different_extra_values(self, collector):
        """Test that assets with different extra values are tracked separately."""
        mock_context = mock.MagicMock()

        # Same URI but different extra metadata
        collector.add_input_asset(mock_context, uri="s3://bucket/file", asset_extra={"version": "1"})
        collector.add_input_asset(mock_context, uri="s3://bucket/file", asset_extra={"version": "2"})

        collector.add_output_asset(mock_context, uri="s3://output/file", asset_extra={"format": "parquet"})
        collector.add_output_asset(mock_context, uri="s3://output/file", asset_extra={"format": "csv"})

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 2
        assert lineage.inputs[0].asset.uri == "s3://bucket/file"
        assert lineage.inputs[0].asset.extra == {"version": "1"}
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[0].context == mock_context
        assert lineage.inputs[1].asset.uri == "s3://bucket/file"
        assert lineage.inputs[1].asset.extra == {"version": "2"}
        assert lineage.inputs[1].count == 1
        assert lineage.inputs[1].context == mock_context

        assert len(lineage.outputs) == 2
        assert lineage.outputs[0].asset.uri == "s3://output/file"
        assert lineage.outputs[0].asset.extra == {"format": "parquet"}
        assert lineage.outputs[0].count == 1
        assert lineage.outputs[0].context == mock_context
        assert lineage.outputs[1].asset.uri == "s3://output/file"
        assert lineage.outputs[1].asset.extra == {"format": "csv"}
        assert lineage.outputs[1].count == 1
        assert lineage.outputs[1].context == mock_context

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3.0+")
    def test_add_asset_max_limit_af3(self, collector):
        """Test that asset operations respect maximum limit."""
        mock_context = mock.MagicMock()
        max_limit = 100
        added_assets = max_limit + 50

        # Limitation on collected assets was added in AF3 #45798
        expected_number = max_limit

        # Add more than max allowed inputs
        for i in range(added_assets):
            collector.add_input_asset(mock_context, uri=f"s3://bucket/input-{i}")

        # Add more than max allowed outputs
        for i in range(added_assets):
            collector.add_output_asset(mock_context, uri=f"s3://bucket/output-{i}")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == expected_number
        assert len(lineage.outputs) == expected_number

    @pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Test requires < Airflow 3.0")
    def test_add_asset_max_limit_af2(self, collector):
        """Test that asset operations do not respect maximum limit."""
        mock_context = mock.MagicMock()
        max_limit = 100
        added_assets = max_limit + 50

        # Limitation on collected assets was added in AF3 #45798
        expected_number = added_assets

        # Add more than max allowed inputs
        for i in range(added_assets):
            collector.add_input_asset(mock_context, uri=f"s3://bucket/input-{i}")

        # Add more than max allowed outputs
        for i in range(added_assets):
            collector.add_output_asset(mock_context, uri=f"s3://bucket/output-{i}")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == expected_number
        assert len(lineage.outputs) == expected_number


class TestEdgeCases:
    """Test edge cases and error conditions to ensure collector never fails."""

    @pytest.mark.parametrize("uri", ["", None])
    def test_invalid_uri_none(self, collector, uri):
        """Test handling of None URI - should not raise."""
        mock_context = mock.MagicMock()

        # Should not raise exceptions
        collector.add_input_asset(mock_context, uri=uri)
        collector.add_output_asset(mock_context, uri=uri)

        # Collector should handle gracefully and not collect invalid URIs
        assert not collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 0
        assert len(lineage.outputs) == 0
        assert len(lineage.extra) == 0

    def test_malformed_uri(self, collector):
        """Test handling of malformed URIs - should not raise."""
        mock_context = mock.MagicMock()

        # Various malformed URIs should not cause crashes
        collector.add_input_asset(mock_context, uri="not-a-valid-uri")
        collector.add_input_asset(mock_context, uri="://missing-scheme")
        collector.add_input_asset(mock_context, uri="scheme:")
        collector.add_output_asset(mock_context, uri="//no-scheme")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 3
        assert lineage.inputs[0].asset.uri == "not-a-valid-uri"
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[0].context == mock_context
        assert lineage.inputs[1].asset.uri == "://missing-scheme"
        assert lineage.inputs[1].count == 1
        assert lineage.inputs[1].context == mock_context
        assert lineage.inputs[2].asset.uri == "scheme:/"
        assert lineage.inputs[2].count == 1
        assert lineage.inputs[2].context == mock_context

        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].asset.uri == "//no-scheme"
        assert lineage.outputs[0].count == 1
        assert lineage.outputs[0].context == mock_context

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3.0+")
    def test_very_long_uri(self, collector):
        """Test handling of very long URIs - 1000 chars OK, 2000 chars raises ValueError."""
        mock_context = mock.MagicMock()

        # Create very long URI (1000 chars - should work)
        long_path = "a" * 1000
        long_uri = f"s3://bucket/{long_path}"

        # Create too long URI (2000 chars - should raise)
        too_long_uri = f"s3://bucket/{long_path * 2}"

        collector.add_input_asset(mock_context, uri=long_uri)

        # Too long URI should raise ValueError
        with pytest.raises(ValueError, match="Asset name cannot exceed"):
            collector.add_output_asset(mock_context, uri=too_long_uri)

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].asset.uri == long_uri
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[0].context == mock_context

        assert len(lineage.outputs) == 0
        assert len(lineage.extra) == 0

    def test_none_context(self, collector):
        """Test handling of None context - should not raise."""
        # Should not raise exceptions
        collector.add_input_asset(None, uri="s3://bucket/input")
        collector.add_output_asset(None, uri="s3://bucket/output")
        collector.add_extra(None, "key", "value")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].asset.uri == "s3://bucket/input"
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[0].context is None

        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].asset.uri == "s3://bucket/output"
        assert lineage.outputs[0].count == 1
        assert lineage.outputs[0].context is None

        assert len(lineage.extra) == 1
        assert lineage.extra[0].key == "key"
        assert lineage.extra[0].value == "value"
        assert lineage.extra[0].count == 1
        assert lineage.extra[0].context is None

    def test_special_characters_in_extra_key(self, collector):
        """Test that extra keys with special characters work."""
        mock_context = mock.MagicMock()

        collector.add_extra(mock_context, "key-with-dashes", {"data": "value"})
        collector.add_extra(mock_context, "key.with.dots", {"data": "value"})
        collector.add_extra(mock_context, "key_with_underscores", {"data": "value"})
        collector.add_extra(mock_context, "key/with/slashes", {"data": "value"})
        collector.add_extra(mock_context, "key:with:colons", {"data": "value"})

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 5
        assert lineage.extra[0].key == "key-with-dashes"
        assert lineage.extra[1].key == "key.with.dots"
        assert lineage.extra[2].key == "key_with_underscores"
        assert lineage.extra[3].key == "key/with/slashes"
        assert lineage.extra[4].key == "key:with:colons"

    def test_unicode_in_extra_key_and_value(self, collector):
        """Test that unicode characters in extra work correctly."""
        mock_context = mock.MagicMock()

        collector.add_extra(mock_context, "clé_française", {"données": "valeur"})
        collector.add_extra(mock_context, "中文键", {"中文": "值"})
        collector.add_extra(mock_context, "مفتاح", {"بيانات": "قيمة"})
        collector.add_extra(mock_context, "emoji_🚀", {"status": "✅"})

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 4
        assert lineage.extra[0].key == "clé_française"
        assert lineage.extra[0].value == {"données": "valeur"}
        assert lineage.extra[0].count == 1
        assert lineage.extra[0].context == mock_context

        assert lineage.extra[1].key == "中文键"
        assert lineage.extra[1].value == {"中文": "值"}
        assert lineage.extra[1].count == 1
        assert lineage.extra[1].context == mock_context

        assert lineage.extra[2].key == "مفتاح"
        assert lineage.extra[2].value == {"بيانات": "قيمة"}
        assert lineage.extra[2].count == 1
        assert lineage.extra[2].context == mock_context

        assert lineage.extra[3].key == "emoji_🚀"
        assert lineage.extra[3].value == {"status": "✅"}
        assert lineage.extra[3].count == 1
        assert lineage.extra[3].context == mock_context

    def test_very_large_extra_value(self, collector):
        """Test that large extra values are handled."""
        mock_context = mock.MagicMock()

        # Create a large value
        large_value = {"data": "x" * 10000, "list": list(range(1000))}

        collector.add_extra(mock_context, "large_key", large_value)

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 1
        assert lineage.extra[0].key == "large_key"
        assert lineage.extra[0].value == large_value
        assert lineage.extra[0].count == 1
        assert lineage.extra[0].context == mock_context

        assert len(lineage.inputs) == 0
        assert len(lineage.outputs) == 0

    def test_deeply_nested_extra_value(self, collector):
        """Test that deeply nested data structures in extra are handled."""
        mock_context = mock.MagicMock()

        # Create deeply nested structure
        nested_value = {"level1": {"level2": {"level3": {"level4": {"level5": {"data": "deep"}}}}}}

        collector.add_extra(mock_context, "nested", nested_value)

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 1
        assert lineage.extra[0].key == "nested"
        assert lineage.extra[0].value == nested_value
        assert lineage.extra[0].count == 1
        assert lineage.extra[0].context == mock_context

        assert len(lineage.inputs) == 0
        assert len(lineage.outputs) == 0

    def test_extra_value_with_various_types(self, collector):
        """Test that extra can handle various data types."""
        mock_context = mock.MagicMock()

        collector.add_extra(mock_context, "string", "text")
        collector.add_extra(mock_context, "integer", 42)
        collector.add_extra(mock_context, "float", 3.14)
        collector.add_extra(mock_context, "boolean", True)
        collector.add_extra(mock_context, "list", [1, 2, 3])
        collector.add_extra(mock_context, "dict", {"key": "value"})
        collector.add_extra(mock_context, "null", None)

        assert collector.has_collected

        # None value should not be collected (based on validation)
        lineage = collector.collected_assets
        assert len(lineage.extra) == 6  # None is filtered out

        assert lineage.extra[0].key == "string"
        assert lineage.extra[0].value == "text"
        assert lineage.extra[0].count == 1

        assert lineage.extra[1].key == "integer"
        assert lineage.extra[1].value == 42
        assert lineage.extra[1].count == 1

        assert lineage.extra[2].key == "float"
        assert lineage.extra[2].value == 3.14
        assert lineage.extra[2].count == 1

        assert lineage.extra[3].key == "boolean"
        assert lineage.extra[3].value is True
        assert lineage.extra[3].count == 1

        assert lineage.extra[4].key == "list"
        assert lineage.extra[4].value == [1, 2, 3]
        assert lineage.extra[4].count == 1

        assert lineage.extra[5].key == "dict"
        assert lineage.extra[5].value == {"key": "value"}
        assert lineage.extra[5].count == 1

        assert len(lineage.inputs) == 0
        assert len(lineage.outputs) == 0

    def test_non_json_serializable_value_in_extra(self, collector):
        """Test that non-JSON-serializable values are handled gracefully."""
        mock_context = mock.MagicMock()

        # Create a non-serializable object
        class CustomObject:
            def __str__(self):
                return "custom_object"

        # Should not raise - collector should handle via str conversion or skip
        collector.add_extra(mock_context, "custom_key", CustomObject())

        # May or may not be collected depending on implementation
        lineage = collector.collected_assets
        # Just verify it doesn't crash
        assert isinstance(lineage.extra, list)
        assert len(lineage.inputs) == 0
        assert len(lineage.outputs) == 0

    def test_empty_asset_extra(self, collector):
        """Test that empty asset_extra is handled correctly."""
        mock_context = mock.MagicMock()

        collector.add_input_asset(mock_context, uri="s3://bucket/file", asset_extra={})
        collector.add_output_asset(mock_context, uri="s3://bucket/file", asset_extra={})

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].asset.extra == {}
        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].asset.extra == {}

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3.0+")
    def test_asset_with_all_optional_parameters(self, collector):
        """Test asset creation with all optional parameters provided."""
        mock_context = mock.MagicMock()

        collector.add_input_asset(
            mock_context,
            uri="s3://bucket/file",
            name="custom-name",
            group="custom-group",
            asset_extra={"key": "value"},
        )

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].asset.uri == "s3://bucket/file"
        assert lineage.inputs[0].asset.name == "custom-name"
        assert lineage.inputs[0].asset.group == "custom-group"
        assert lineage.inputs[0].asset.extra == {"key": "value"}

    def test_rapid_repeated_calls(self, collector):
        """Test that rapid repeated calls don't cause issues."""
        mock_context = mock.MagicMock()

        # Simulate rapid repeated calls
        for _ in range(50):
            collector.add_input_asset(mock_context, uri="s3://bucket/file")
            collector.add_output_asset(mock_context, uri="s3://bucket/output")
            collector.add_extra(mock_context, "key", "value")

        assert collector.has_collected

        lineage = collector.collected_assets
        # Should have counted properly
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].count == 50
        assert len(lineage.outputs) == 1
        assert lineage.outputs[0].count == 50
        assert len(lineage.extra) == 1
        assert lineage.extra[0].count == 50

    def test_mixed_valid_invalid_operations(self, collector):
        """Test mixing valid and invalid operations."""
        mock_context = mock.MagicMock()

        # Mix valid and invalid calls
        collector.add_input_asset(mock_context, uri="s3://bucket/valid")
        collector.add_input_asset(mock_context, uri=None)  # Invalid - should not be collected
        collector.add_input_asset(mock_context, uri="")  # Invalid - should not be collected
        collector.add_input_asset(mock_context, uri="s3://bucket/another-valid")

        collector.add_extra(mock_context, "valid_key", "valid_value")
        collector.add_extra(mock_context, "", "invalid_key")  # Invalid key - should not be collected
        collector.add_extra(mock_context, "another_key", "another_value")

        assert collector.has_collected

        # Should collect only valid items
        lineage = collector.collected_assets
        assert len(lineage.inputs) == 2
        assert lineage.inputs[0].asset.uri == "s3://bucket/valid"
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[0].context == mock_context
        assert lineage.inputs[1].asset.uri == "s3://bucket/another-valid"
        assert lineage.inputs[1].count == 1
        assert lineage.inputs[1].context == mock_context

        assert len(lineage.extra) == 2
        assert lineage.extra[0].key == "valid_key"
        assert lineage.extra[0].value == "valid_value"
        assert lineage.extra[0].count == 1
        assert lineage.extra[1].key == "another_key"
        assert lineage.extra[1].value == "another_value"
        assert lineage.extra[1].count == 1

        assert len(lineage.outputs) == 0

    def test_collector_collected_assets_called_multiple_times(self, collector):
        """Test that collected_assets property can be called multiple times."""
        mock_context = mock.MagicMock()

        collector.add_input_asset(mock_context, uri="s3://bucket/file")

        assert collector.has_collected

        # Call multiple times - should return same data
        lineage1 = collector.collected_assets
        lineage2 = collector.collected_assets
        lineage3 = collector.collected_assets

        assert lineage1.inputs == lineage2.inputs == lineage3.inputs
        assert len(lineage1.inputs) == 1
        assert lineage1.inputs[0].asset.uri == "s3://bucket/file"
        assert lineage1.inputs[0].count == 1
        assert lineage1.inputs[0].context == mock_context

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3.0+")
    def test_empty_name_and_group(self, collector):
        """Test that empty strings for name and group are handled."""
        mock_context = mock.MagicMock()

        # Empty strings for optional parameters
        collector.add_input_asset(mock_context, uri="s3://bucket/file", name="", group="")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.inputs) == 1
        assert lineage.inputs[0].asset.uri == "s3://bucket/file"
        assert lineage.inputs[0].asset.name == "s3://bucket/file"
        assert lineage.inputs[0].asset.group == "asset"
        assert lineage.inputs[0].count == 1
        assert lineage.inputs[0].context == mock_context

        assert len(lineage.outputs) == 0
        assert len(lineage.extra) == 0

    def test_extremely_long_extra_key(self, collector):
        """Test that extremely long extra keys are handled."""
        mock_context = mock.MagicMock()

        long_key = "k" * 10000
        collector.add_extra(mock_context, long_key, "value")

        assert collector.has_collected

        lineage = collector.collected_assets
        assert len(lineage.extra) == 1
        assert lineage.extra[0].key == long_key
        assert lineage.extra[0].value == "value"
        assert lineage.extra[0].count == 1
        assert lineage.extra[0].context == mock_context

        assert len(lineage.inputs) == 0
        assert len(lineage.outputs) == 0



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from airflow.providers.openlineage.sqlparser import get_openlineage_facets_with_sql

else:
    try:
        from airflow.providers.openlineage.sqlparser import get_openlineage_facets_with_sql
    except ImportError:

        def get_openlineage_facets_with_sql(
            hook,
            sql: str | list[str],
            conn_id: str,
            database: str | None,
        ):
            try:
                from airflow.providers.openlineage.sqlparser import SQLParser
            except ImportError:
                log.debug("SQLParser could not be imported from OpenLineage provider.")
                return None

            try:
                from airflow.providers.openlineage.utils.utils import should_use_external_connection

                use_external_connection = should_use_external_connection(hook)
            except ImportError:
                # OpenLineage provider release < 1.8.0 - we always use connection
                use_external_connection = True

            connection = hook.get_connection(conn_id)
            try:
                database_info = hook.get_openlineage_database_info(connection)
            except AttributeError:
                log.debug("%s has no database info provided", hook)
                database_info = None

            if database_info is None:
                return None

            try:
                sql_parser = SQLParser(
                    dialect=hook.get_openlineage_database_dialect(connection),
                    default_schema=hook.get_openlineage_default_schema(),
                )
            except AttributeError:
                log.debug("%s failed to get database dialect", hook)
                return None

            try:
                sqlalchemy_engine = hook.get_sqlalchemy_engine()
            except Exception as e:
                log.debug("Failed to get sql alchemy engine: %s", e)
                sqlalchemy_engine = None

            operator_lineage = sql_parser.generate_openlineage_metadata_from_sql(
                sql=sql,
                hook=hook,
                database_info=database_info,
                database=database,
                sqlalchemy_engine=sqlalchemy_engine,
                use_connection=use_external_connection,
            )

            return operator_lineage


__all__ = ["get_openlineage_facets_with_sql"]



#!/usr/bin/env python3
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import pytest


def test_all_compat_imports_work():
    """
    Test that all items in _IMPORT_MAP can be successfully imported.

    For each item, validates that at least one of the specified import paths works,
    ensuring the fallback mechanism is functional.
    """
    from airflow.providers.common.compat import sdk

    failed_imports = []

    for name in sdk.__all__:
        try:
            obj = getattr(sdk, name)
            assert obj is not None, f"{name} imported as None"
        except (ImportError, AttributeError) as e:
            failed_imports.append((name, str(e)))

    if failed_imports:
        error_msg = "The following imports failed:\n"
        for name, error in failed_imports:
            error_msg += f"  - {name}: {error}\n"
        pytest.fail(error_msg)


def test_invalid_import_raises_attribute_error():
    """Test that importing non-existent attribute raises AttributeError."""
    from airflow.providers.common.compat import sdk

    with pytest.raises(AttributeError, match="has no attribute 'NonExistentClass'"):
        _ = sdk.NonExistentClass



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from importlib import metadata
from unittest.mock import patch

import pytest

from airflow.providers.common.compat.check import require_provider_version
from airflow.providers.common.compat.sdk import AirflowOptionalProviderFeatureException


def test_decorator_usage_without_parentheses():
    with pytest.raises(TypeError):

        @require_provider_version
        def dummy_function():
            pass


def test_empty_provider_name_and_version():
    with pytest.raises(ValueError, match="decorator must be used with two arguments"):

        @require_provider_version("", "")
        def dummy_function():
            pass


def test_invalid_provider_name():
    expected_error = (
        "Full `provider_name` must be provided starting with 'apache-airflow-providers-', "
        "got `invalid-provider-name`."
    )
    with pytest.raises(ValueError, match=expected_error):

        @require_provider_version("invalid-provider-name", "1.0.0")
        def dummy_function():
            pass


@patch("importlib.metadata.version", return_value="0.9.9")
def test_provider_version_lower_than_required(mock_version):
    @require_provider_version("apache-airflow-providers-nonexistingprovidermock", "1.0.0")
    def dummy_function():
        return "Function Executed"

    with pytest.raises(
        AirflowOptionalProviderFeatureException,
        match=r"Provider's `apache-airflow-providers-nonexistingprovidermock` "
        r"version `0.9.9` is lower than required `1.0.0`",
    ):
        dummy_function()


@patch("importlib.import_module", side_effect=ImportError)
@patch("importlib.metadata.version", side_effect=metadata.PackageNotFoundError)
def test_provider_not_installed(mock_version, mock_import):
    @require_provider_version("apache-airflow-providers-nonexistingprovidermock", "1.0.0")
    def dummy_function():
        return "Function Executed"

    with pytest.raises(
        AirflowOptionalProviderFeatureException,
        match=r"Provider `apache-airflow-providers-nonexistingprovidermock` not found or has no version",
    ):
        dummy_function()


@patch("importlib.metadata.version", return_value="2.0.0")
def test_provider_version_ok(mock_version):
    @require_provider_version("apache-airflow-providers-nonexistingprovidermock", "1.0.0")
    def dummy_function():
        return "Function Executed"

    result = dummy_function()
    assert result == "Function Executed"


@patch("importlib.import_module", return_value=type("module", (), {"__version__": "1.5.0"}))
@patch("importlib.metadata.version", side_effect=metadata.PackageNotFoundError)
def test_provider_dynamic_import(mock_version, mock_import):
    @require_provider_version("apache-airflow-providers-nonexistingprovidermock", "1.0.0")
    def dummy_function():
        return "Function Executed"

    result = dummy_function()
    assert result == "Function Executed"
